/*-
 * Copyright (c) 2006 John Birrell jb@freebsd.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/types.h>
#include <libelf.h>
#include <stdio.h>
#include <unistd.h>

#include "_libelf.h"

int
elf_dump_argv(int argc, char **argv)
{
	char	*name = NULL;
	char	*out_file = NULL;
	int	f_usage = 0;
	int	i;
	int	j;
	int	ret = 0;
	u_int	flags = 0;

	for (i = 1; i < argc && argv[i][0] == '-'; i++) {
		for (j = 1;  j > 0 && argv[i][j] != '\0'; j++) {
			switch (argv[i][j]) {
			case 'c':
				flags |= ELF_DUMP_SHDR;
				break;
			case 'C':
				flags |= ELF_DUMP_DEMANGLE;
				break;
			case 'd':
				flags |= ELF_DUMP_DYN;
				break;
			case 'e':
				flags |= ELF_DUMP_EHDR;
				break;
			case 'g':
				flags |= ELF_DUMP_GROUP;
				break;
			case 'G':
				flags |= ELF_DUMP_GOT;
				break;
			case 'h':
				flags |= ELF_DUMP_HASH;
				break;
			case 'H':
				flags |= ELF_DUMP_CAP;
				break;
			case 'i':
				flags |= ELF_DUMP_INTERP;
				break;
			case 'k':
				flags |= ELF_DUMP_CHECKSUM;
				break;
			case 'l':
				flags |= ELF_DUMP_LONGNAME;
				break;
			case 'm':
				flags |= ELF_DUMP_MOVE;
				break;
			case 'n':
				flags |= ELF_DUMP_NOTE;
				break;
			case 'N':
				if (i == argc - 1) {
					fprintf(stderr,"Name not specified with -n\n");
					f_usage = 1;
				} else
					name = argv[++i];

				/* Leave the inner loop. */
				j = -1;
				break;
			case 'p':
				flags |= ELF_DUMP_PHDR;
				break;
			case 'r':
				flags |= ELF_DUMP_RELOC;
				break;
			case 's':
				flags |= ELF_DUMP_SYMBOLS;
				break;
			case 'u':
				flags |= ELF_DUMP_UNWIND;
				break;
			case 'v':
				flags |= ELF_DUMP_VERSIONS;
				break;
			case 'w':
				if (i == argc - 1) {
					fprintf(stderr,"File name not specified with -w\n");
					f_usage = 1;
				} else {
					out_file = argv[++i];
					unlink(out_file);
				}

				/* Leave the inner loop. */
				j = -1;
				break;
			case 'y':
				flags |= ELF_DUMP_SYMINFO;
				break;
			case '?':
				f_usage = 1;
				break;
			default:
				printf("invalid option '%c'\n",argv[i][j]);
				f_usage = 1;
				break;
			}
		}
	}

	if (f_usage || argc == 0 || i == argc) {
		fprintf(stderr, "Usage: elfdump [-cCdegGhHiklmnprsuvy] [-N name] [-w file] filename\n");
		return (1);
	}

	/* If no options were specified, dump everything. */
	if (!flags)
		flags = ~0;

	while (i < argc && ret == 0)
		ret = elf_dump(argv[i++], name, out_file, flags);

	return (ret);
}

