/*-
 * Copyright (c) 2006 John Birrell jb@freebsd.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#if LIBELF_DUMP_NN == 32
#define	_libelf_dump	_libelf_dump32
#define ELF_EHDR	Elf32_Ehdr
#define ELF_GETEHDR	elf32_getehdr
#else
#define	_libelf_dump	_libelf_dump64
#define ELF_EHDR	Elf64_Ehdr
#define ELF_GETEHDR	elf64_getehdr
#endif

int
_libelf_dump(Elf *e, __unused const char *name, __unused const char *outfile, unsigned int flags)
{
	ELF_EHDR *ehdr;
	const char *s;
	char s1[16];

	if ((ehdr = ELF_GETEHDR(e)) == NULL)
		return (1);

	if ((flags & ELF_DUMP_EHDR) != 0) {
		printf("\nelf header:\n\n");
		printf("\te_ident:");
		if ((s = ei_class_str(ehdr->e_ident[EI_CLASS])) == NULL) {
			snprintf(s1, sizeof(s1), "%u", ehdr->e_ident[EI_CLASS]);
			s = s1;
		}
		printf(" %s", s);
		if ((s = ei_data_str(ehdr->e_ident[EI_DATA])) == NULL) {
			snprintf(s1, sizeof(s1), "%u", ehdr->e_ident[EI_DATA]);
			s = s1;
		}
		printf(" %s", s);
		if ((s = ei_osabi_str(ehdr->e_ident[EI_OSABI])) == NULL) {
			snprintf(s1, sizeof(s1), "%u", ehdr->e_ident[EI_OSABI]);
			s = s1;
		}
		printf(" %s\n", s);
		if ((s = et_str(ehdr->e_type)) == NULL) {
			snprintf(s1, sizeof(s1), "%hu", ehdr->e_type);
			s = s1;
		}
		printf("\te_type: %s\n", s);
		if ((s = em_str(ehdr->e_machine)) == NULL) {
			snprintf(s1, sizeof(s1), "%hu", ehdr->e_machine);
			s = s1;
		}
		printf("\te_machine: %s\n", s);
		if ((s = ev_str(ehdr->e_version)) == NULL) {
			snprintf(s1, sizeof(s1), "%u", ehdr->e_version);
			s = s1;
		}
		printf("\te_version: %s\n", s);
		printf("\te_entry: 0x%jx\n", (uint64_t) ehdr->e_entry);
		printf("\te_phoff: %ju\n", (uint64_t) ehdr->e_phoff);
		printf("\te_shoff: %ju\n", (uint64_t) ehdr->e_shoff);
		printf("\te_flags: %u\n", ehdr->e_flags);
		printf("\te_ehsize: %hu\n", ehdr->e_ehsize);
		printf("\te_phentsize: %hu\n", ehdr->e_phentsize);
		printf("\te_phnum: %hu\n", ehdr->e_phnum);
		printf("\te_shentsize: %hu\n", ehdr->e_shentsize);
		printf("\te_shnum: %hu\n", ehdr->e_shnum);
		printf("\te_shstrndx: %hu\n", ehdr->e_shstrndx);
	}
	return 0;
}

#undef _libelf_dump
#undef ELF_EHDR
#undef ELF_GETEHDR

